using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Npgsql;

namespace Shape2Pgsql
{
	public partial class Form1 : Form
	{
		SharpMap.Data.Providers.ShapeFile shp;
		SharpMap.Map map;

		public Form1()
		{
			InitializeComponent();
		}

		#region UI control and other boring stuff

		private void textBox1_TextChanged(object sender, EventArgs e)
		{
			//Check if a valid shapefile is in the textbox
			bool isValidFile = (tbShapefile.Text.ToLower().EndsWith(".shp") && System.IO.File.Exists(tbShapefile.Text));
			btnPreview.Enabled = isValidFile;
			groupBox2.Enabled = isValidFile;
			if (shp != null) shp.Dispose();
			if (isValidFile) //Initialize shapefile provider
			{
				shp = new SharpMap.Data.Providers.ShapeFile(tbShapefile.Text, chbSaveIndex.Checked);
				CreateMap();
				pictureBox1.Cursor = Cursors.Cross;
			}
			else
			{
				shp = null;
				pictureBox1.Cursor = Cursors.No;
				pictureBox1.Image = null;
			}
		}

		/// <summary>
		/// Shows filedialog to browse for a shapefile
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnSelectShp_Click(object sender, EventArgs e)
		{
			OpenFileDialog ofd = new OpenFileDialog();
			if (System.IO.File.Exists(tbShapefile.Text))
				ofd.FileName = tbShapefile.Text;
			ofd.Filter = "Shapefile (*.shp)|*.shp";
			if (ofd.ShowDialog() != DialogResult.OK)
				return;
			tbShapefile.Text = ofd.FileName;
		}


		private void btnPreview_Click_1(object sender, EventArgs e)
		{
			map.Size = pictureBox1.Size;
			pictureBox1.Image = map.GetMap();
		}
						
		private void chbSRID_CheckedChanged(object sender, EventArgs e)
		{
			tbSRID.Enabled = chbSRID.Checked;
		}

		/// <summary>
		/// Zooms the map preview (left-click: zoom in, right-click: zoom out)
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void pictureBox1_MouseClick_1(object sender, MouseEventArgs e)
		{
			if (!btnPreview.Enabled)
				return;
			if (e.Button == MouseButtons.Left)
			{
				map.Center = map.ImageToWorld(new PointF(e.X, e.Y));
				map.Zoom *= 0.5;
			}
			else if (e.Button == MouseButtons.Right)
			{
				map.Center = map.ImageToWorld(new PointF(pictureBox1.Width - e.X, pictureBox1.Height - e.Y));
				map.Zoom *= 2;
			}
			else return;
			pictureBox1.Image = map.GetMap();

		}

		private void pictureBox1_MouseMove(object sender, MouseEventArgs e)
		{
			if (map != null)
				lbMousePosition.Text = map.ImageToWorld(new PointF(e.X, e.Y)).ToString();
		}

		private void pictureBox1_MouseLeave(object sender, EventArgs e)
		{
			lbMousePosition.Text = "";
		}

		private void btnAbout_Click(object sender, EventArgs e)
		{
			new frmAbout().ShowDialog();
		}

		/// <summary>
		/// Test the connection to the database
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnTestConnection_Click(object sender, EventArgs e)
		{
			string ConnStr = "Server=" + tbServer.Text +
							 ";Port=" + tbPort.Text +
							 ";User Id=" + tbUsername.Text +
							 ";Password=" + tbPassword.Text +
							 ";Database=" + tbDatabase.Text + ";";
			try
			{
				NpgsqlConnection conn = new NpgsqlConnection(ConnStr);
				conn.Open();
				NpgsqlCommand command = new NpgsqlCommand();
				command.Connection = conn;
				//Select the current PostGIS version for test
				string strSQL = "select postgis_full_version();";
				command.CommandText = strSQL;
				string result = command.ExecuteScalar().ToString();
				MessageBox.Show("Server connection OK (" + result);
			}
			catch (System.Exception ex)
			{
				MessageBox.Show("Connection failed: " + ex.Message);
			}
		}

#endregion

		/// <summary>
		/// Initializes the map with the currently selected shapefile
		/// </summary>
		private void CreateMap()
		{
			//Initialize the map
			map = new SharpMap.Map(pictureBox1.Size);
			//Setup a shapefile vectorlayer with some default styles
			SharpMap.Layers.VectorLayer layer = new SharpMap.Layers.VectorLayer("preview layer");
			layer.DataSource = shp;
			layer.Style.Line = new Pen(Color.Blue);
			layer.Style.Outline = new Pen(Color.Black);
			layer.Style.EnableOutline = true;
			layer.Style.Fill = Brushes.Green;
			layer.Style.Symbol = new Bitmap(System.Reflection.Assembly.GetExecutingAssembly().GetManifestResourceStream("Shape2Pgsql.PointSymbol.bmp"));
			map.Layers.Add(layer);
			map.ZoomToExtents();
		}		

		/// <summary>
		/// Uploads the shapefile to the PostGIS database (this is where all the fun takes place :-)
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnUpload_Click(object sender, EventArgs e)
		{
			string ConnStr = "Server=" + tbServer.Text +
							 ";Port=" + tbPort.Text +
							 ";User Id=" + tbUsername.Text +
							 ";Password=" + tbPassword.Text +
							 ";Database=" + tbDatabase.Text + ";";
			string tablename = shp.Filename.Substring(shp.Filename.LastIndexOf('\\')+1,
													  shp.Filename.LastIndexOf('.')-shp.Filename.LastIndexOf('\\')-1);
			shp.Open();
			int counter = 0;
			int SRID = (chbSRID.Checked?int.Parse(tbSRID.Text):-1);
			using (NpgsqlConnection conn = new NpgsqlConnection(ConnStr))
			{
				conn.Open();
				List<uint> indexes = shp.GetObjectIDsInView(shp.GetExtents());
				
				NpgsqlCommand command = new NpgsqlCommand();
				command.Connection = conn;
				
				//Retrieve the first feature to determine geometrytype and columns
				SharpMap.Data.FeatureDataRow geom = shp.GetFeature(indexes[0]);
				DataColumnCollection columns = geom.Table.Columns;

				//If we want to replace the table, try to drop it first
				if (chbReplace.Checked)
				{
					try
					{
						command.CommandText = "DROP TABLE \"" + tablename + "\";";
						command.ExecuteNonQuery();
					}
					catch { }
				}

				//Create table with a unique object ID column 'oid'
				command.CommandText = "CREATE TABLE \"" + tablename + "\" (oid serial PRIMARY KEY";
				foreach(DataColumn col in columns)
					command.CommandText += ", \"" + col.ColumnName+ "\" " + Type2PgType(col.DataType);
				command.CommandText += ");";
				command.ExecuteNonQuery();

				//Add geometry column
				command.CommandText = "SELECT AddGeometryColumn('','" + tablename + "','geom','" + SRID.ToString() + "','GEOMETRY',2);";
				command.ExecuteNonQuery();

				//Select all indexes in shapefile, loop through each feature and insert them one-by-one
				foreach (uint idx in indexes)
				{				
						//Get feature from shapefile
						SharpMap.Data.FeatureDataRow feature = shp.GetFeature(idx);
						//Create insert script
						string strSQL = "INSERT INTO \"" + tablename + "\"";
						
						strSQL += " (\"geom\"";
						foreach (DataColumn col in feature.Table.Columns)
							strSQL += ",\"" + col.ColumnName + "\"";

						strSQL += ") VALUES (GeomFromWKB(:geom,:srid)";
						foreach (DataColumn col in feature.Table.Columns)
							strSQL += ",:" + col.ColumnName;

						strSQL += ");";
						command.CommandText = strSQL;						

						//Add geometry to row
						command.Parameters.Add(":geom", NpgsqlTypes.NpgsqlDbType.Bytea);
						command.Parameters[":geom"].Value = feature.Geometry.AsBinary(); //Add the geometry as Well-Known Binary
						command.Parameters.Add(":srid", NpgsqlTypes.NpgsqlDbType.Integer);
						command.Parameters[":srid"].Value = SRID; //Set the SRID of the geometry
						//Add datacolumns to row
						foreach (DataColumn col in feature.Table.Columns)
						{
							command.Parameters.Add(col.ColumnName, Type2NpgsqlType(col.DataType));
							command.Parameters[col.ColumnName].Value = feature[col];
						}
						command.ExecuteNonQuery();
						counter++;			
				}
				//Build spatial index
				command.CommandText = "CREATE INDEX \"" + tablename + "_sidx\" ON \"" + tablename + "\" USING gist (geom);";
				command.ExecuteNonQuery();

				conn.Close();
			}
			shp.Close();
			MessageBox.Show("Uploaded " + counter.ToString() + " features to table " + tablename);
		}


		#region NGPSQL helper-functions

		/// <summary>
		/// Returns the name of the PostGreSQL/PostGIS datatype based on a .NET datatype
		/// </summary>
		/// <param name="t"></param>
		/// <returns></returns>
		private string Type2PgType(Type t)
		{
			switch(t.ToString())
			{
				case "System.Boolean": return "bool";
				case "System.Single": return "float4";
				case "System.Double": return "float8";
				case "System.Int32": return "int4";
				case "System.DateTime": return "date";
				case "System.Byte[]": return "binary";
				case "System.String": return "varchar";
				case "SharpMap.Geometries.Polygon": return "MULTIPOLYGON";
				case "SharpMap.Geometries.MultiPolygon": return "MULTIPOLYGON";
				case "SharpMap.Geometries.LineString": return "MULTILINESTRING";
				case "SharpMap.Geometries.MultiLineString": return "MULTILINESTRING";
				case "SharpMap.Geometries.Point": return "MULTIPOINT";
				case "SharpMap.Geometries.MultiPoint": return "MULTIPOINT";
				default:
					throw(new NotSupportedException("Unsupported datatype '" + t.Name + "' found in Shapefile"));
			}
		}
		/// <summary>
		/// Returns the corresponding ngpsql datatype based on a .NET datatype
		/// </summary>
		private NpgsqlTypes.NpgsqlDbType Type2NpgsqlType(Type t)
		{
			switch (t.ToString())
			{
				case "System.Boolean": return NpgsqlTypes.NpgsqlDbType.Boolean;
				case "System.Single": return NpgsqlTypes.NpgsqlDbType.Real;
				case "System.Double": return NpgsqlTypes.NpgsqlDbType.Double;
				case "System.Int32": return NpgsqlTypes.NpgsqlDbType.Integer;
				case "System.DateTime": return NpgsqlTypes.NpgsqlDbType.Date;
				case "System.Byte[]": return NpgsqlTypes.NpgsqlDbType.Bytea;
				case "System.String": return NpgsqlTypes.NpgsqlDbType.Varchar;
				default:
					throw (new NotSupportedException("Unsupported datatype '" + t.Name + "' found in Shapefile"));
			}
		}	

		private void Form1_Load(object sender, EventArgs e)
		{
			tbShapefile.Text = @"c:\GIS\ESRI\ESRIDATA\WORLD\cntry92.shp";
			tbPassword.Text = "C#rule";
			tbDatabase.Text = "gisdata";
		}
		#endregion
	}
}